import os

from typing import List

import openai

from llm_foundation import logger
from llm_foundation.prompts import BASIC_TEXT_SYSTEM_PROMPT


openai.api_key = os.environ['OPENAI_API_KEY']


def get_available_models():
    return openai.models.list()


def get_model_names(filters: List[str] = []):
    
    def filter_f(model):
        for filter in filters:
            if filter in model.id:
                return True
        return False    
    
    return [model.id for model in get_available_models() if filter_f(model)]
    
    
def single_text_request(prompt: str, 
                        system_prompt:str = BASIC_TEXT_SYSTEM_PROMPT, 
                        model: str = "gpt4o-mini"):

    completion = openai.ChatCompletion.create(
        model=model,
        # Pre-define conversation messages for the possible roles 
        messages=[
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": prompt}
        ]
        )
    return completion.choices[0].message




from typing import List
from llm_foundation import logger
import llm_foundation.langchain as llm_lc
import llm_foundation.open_ai as llm_openai
from llm_foundation.basic_structs import LMConfig, Provider
    

def get_model_catalog(provider: Provider, name_filters: List[str] = []):
    
    def get_model_names(provider: Provider, filters: List[str] = []):
        match provider:
            case Provider.LC:
                return llm_openai.get_model_names(filters)
            case Provider.OpenAI:
                return llm_openai.get_model_names(filters)
            case Provider.Autogen:
                return llm_openai.get_model_names(filters)            
            case _:
                raise ValueError("Invalid backend provider")

    catalog = {}
    for provider in Provider:
        print(provider, provider.name)
        for model_name in get_model_names(provider, filters=name_filters):
            catalog[f"{provider.name.lower()}_{model_name}"] = model_name

    return catalog        
        

def get_lm(config: LMConfig):
    logger.info(f"Creating lm object for provider {config.provider.name}")
    match config.provider:
        case Provider.Autogen:
            lm = None
        case Provider.LC:
            lm = llm_lc.get_lm(config)
        case _:
            raise ValueError("Invalid backend provider")

    return lm



from abc import ABC
from typing import List, Optional
from langchain_core.messages import ToolMessage, AIMessage
from langchain_core.runnables import Runnable, RunnableConfig
from langchain_core.runnables.utils import Input, Output
from langchain.schema.agent import AgentFinish
from langgraph.prebuilt import ToolInvocation, ToolExecutor

from llm_foundation import logger
from llm_foundation.utils import banner

# def route(result):
#     if isinstance(result, AgentFinish):
#         return result.return_values['output']
#     else:
#         tools = {
#             "search_wikipedia": search_wikipedia, 
#             "get_current_temperature": get_current_temperature,
#         }
#         return tools[result.tool].run(result.tool_input)


class ToolMaster(Runnable[AIMessage, List], ABC):
    
    def __init__(self, available_tools):
        self.tool_executor = ToolExecutor(available_tools)
    
    @banner(text="Pre Call Tool", level=2)
    def pre_call_tool(self, state):
        messages = state["messages"]
        # Based on the continue condition
        # we know the last message involves a function call
        last_message = messages[-1]
        # We construct an ToolInvocation from the function_call
        return last_message

    @banner(text="Post Call Tool", level=2)
    def post_call_tool(self, state, responses):
        tool_messages = []
        for response in responses:
            tool_call_id, tool_name, response = response
            tool_message = ToolMessage(
                content=str(response), name=tool_name, tool_call_id=tool_call_id
            )
            tool_messages.append(tool_message)

        return {
            "last_node": "call_tools",
            "messages": tool_messages,  # We return a list, because this will get added to the existing list of messages
        }

    @banner(text="Call Tool", level=2)
    def _call_tool(self, tool_call_definition):
        action = ToolInvocation(
            tool=tool_call_definition["name"],
            tool_input=tool_call_definition["args"],
        )
        # We call the tool_executor and get back a response
        response = self.tool_executor.invoke(action)
        logger.info(f"Response ({type(response)}): {response}")
        return (tool_call_definition["id"], tool_call_definition["name"], response)

    def call_tools(self, message: AIMessage) -> List:
        responses = []
        for tool_call_definition in message.tool_calls:
            # add tuples of (tool_call_id, tool_name, response)
            responses.append(self._call_tool(tool_call_definition))
        return responses

    @banner(text="Agentic Tool Call", level=1)
    def agentic_tool_call(self, state):
        message = self.pre_call_tool(state)
        responses = self.call_tools(message)
        return self.post_call_tool(state, responses)


    # Runnable implementation
    def invoke(self, input: AIMessage, config: Optional[RunnableConfig] = None) -> List:
        return self.call_tools(input)



import json
import tempfile
import ast

from pathlib import Path
from llm_foundation import logger
from llm_foundation.agent_types import Example, Persona, Role

def test_persona_to_yaml():
    role1 = Role(name="role1", description="Role 1 Description", agent_system_message="Agent System Message")
    role2 = Role(name="role2", description="Role 2 Description", agent_system_message="Agent System Message again")
    roles_dict= {"role1": role1, "role2": role2}
    persona = Persona(name="Test Persona", roles=roles_dict)
    
    with tempfile.TemporaryDirectory() as temp_dir:
        
        filename="test_persona.yaml"

        persona.to_yaml_file(Path(temp_dir), filename)
        
        same_persona = Persona.from_yaml_file(str(Path(temp_dir, filename)))

        # Assertions
        assert persona.name == same_persona.name, "Not the same name"
        assert len(persona.roles) == len(same_persona.roles), "Not the same number of roles"
        
        for role_name, role in persona.roles.items():
            same_role = same_persona.roles.get(role_name)
            assert same_role is not None, f"Role {role_name} not found in deserialized persona"
            assert role.name == same_role.name, f"Role name mismatch for {role_name}"
            assert role.description == same_role.description, f"Role description mismatch for {role_name}"
            assert role.agent_system_message == same_role.agent_system_message, f"Agent system message mismatch for {role_name}"

def test_persona_from_yaml():
    yaml_content = """
    name: Test Persona
    roles:
        role1:
            name: role1
            description: Role 1 Description
            agent_system_message: Agent System Message
            tasks:
                - name: task1
                  description: This is task 1
                  expected_output: This is expected output for task 1
                - name: task2
                  description: This is task 2
                  expected_output: This is expected output for task 2
        role2:
            name: role2
            description: Role 2 Description
            agent_system_message: Agent System Message again
            examples:
                - header: example 1
                  format: text
                  content: This is an example of example 1
                - header: example 2
                  format: json
                  content: |
                    {
                        'header': 'example 2', 
                        'format': 'json', 
                        'content': 'This is an example of example 2'
                    }
            
    """
    
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test_persona.yaml"
        with open(file_path, 'w') as file:
            file.write(yaml_content)
        
        persona = Persona.from_yaml_file(str(file_path))

        # Assertions
        assert persona.name == "Test Persona", "Persona name mismatch"
        assert len(persona.roles) == 2, "Number of roles mismatch"
        
        role1 = persona.roles.get("role1")
        assert role1 is not None, "Role role1 not found"
        assert role1.name == "role1", "Role1 name mismatch"
        assert role1.description == "Role 1 Description", "Role1 description mismatch"
        assert role1.agent_system_message == "Agent System Message", "Role1 agent system message mismatch"
        assert len(role1.tasks) == 2, "Number of tasks in role1 mismatch"
        
        task1 = role1.tasks[0]
        assert task1.name == "task1", "Task1 name mismatch"
        assert task1.description == "This is task 1", "Task1 description mismatch"
        assert task1.expected_output == "This is expected output for task 1", "Task1 expected output mismatch"
        
        task2 = role1.tasks[1]
        assert task2.name == "task2", "Task2 name mismatch"
        assert task2.description == "This is task 2", "Task2 description mismatch"
        assert task2.expected_output == "This is expected output for task 2", "Task2 expected output mismatch"
                
        role2 = persona.roles.get("role2")
        assert role2 is not None, "Role role2 not found"
        assert role2.name == "role2", "Role2 name mismatch"
        assert role2.description == "Role 2 Description", "Role2 description mismatch"
        assert role2.agent_system_message == "Agent System Message again", "Role2 agent system message mismatch"
        assert len(role2.tasks) == 0, "Number of tasks in role2 mismatch"

        assert len(role2.examples) == 2, "Number of examples in role2 mismatch"        
        example1 = role2.examples[0]
        assert example1.header == "example 1", "Example1 header mismatch"
        assert example1.format == "text", "Example1 format mismatch"
        assert example1.content == "This is an example of example 1", "Example1 content mismatch"
        
        example2 = role2.examples[1]
        assert example2.header == "example 2", "Example2 header mismatch"
        assert example2.format == "json", "Example2 format mismatch"
        if example2.format == "json":
            example2_json = ast.literal_eval(example2.content)
            assert example2_json['header'] == "example 2", "Example2 JSON header mismatch"
            assert example2_json['format'] == "json", "Example2 JSON format mismatch"
            assert example2_json['content'] == "This is an example of example 2", "Example2 JSON content mismatch"
        
        # from rich.pretty import pprint
        # from io import StringIO
        # from contextlib import redirect_stdout
        # str_io = StringIO()
        # with redirect_stdout(str_io):
        #     pprint(role1)
        # output = str_io.getvalue()
        # logger.warning(output)


def test_get_examples_as_str_from_role():
    examples = [
        Example(header="example 1", format="json", content='{"key": "value"}'),
        Example(header="example 2", format="json", content='{"another_key": "another_value"}')
    ]
    role = Role(
        name="role_with_examples",
        description="Role with examples",
        agent_system_message="Agent System Message",
        examples=examples
    )
    
    examples_str = role.get_examples_as_str()
    
    logger.info(examples_str)
    
    expected_str = "example 1" + "\n\n" + json.dumps({"key": "value"}, indent=4) + "\n\n" + "example 2" + "\n\n" + json.dumps({"another_key": "another_value"}, indent=4) + "\n\n"
    
    assert examples_str == expected_str, "Examples string mismatch"
    


import os
from typing import List

import openai

from langchain_openai import ChatOpenAI

from llm_foundation import logger
from llm_foundation.basic_structs import LMConfig, ModelType


openai.api_key = os.environ['OPENAI_API_KEY']


def get_model_names(filters: List[str] = []):
    return []


def get_lm(config: LMConfig):
    logger.info(f"Creating {config.type} object for model {config.model}")
    match config.type:
        case ModelType.Simple:
            pass
        case ModelType.Chat:
            lm = ChatOpenAI(
                model=config.model,
                temperature=config.temperature,
                max_tokens=config.max_tokens
            )
        case _:
            pass
        
    logger.info(f"LM object {lm} created")
    return lm



import json
import yaml

from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union
from enum import Enum

from autogen.agentchat import ConversableAgent, GroupChat, AssistantAgent, UserProxyAgent, GroupChatManager
from crewai import Agent, Task as CrewAITask
from pydantic import BaseModel
from pydantic_yaml import parse_yaml_raw_as
from pathlib import Path


from llm_foundation import logger


DEFAULT_TERMINATION_WORD="TERMINATE"

def is_termination_message(msg, termination_word: str = DEFAULT_TERMINATION_WORD) -> bool:
    '''Detects if we should terminate the conversation'''
    
    if isinstance(msg.get("content"), str):
        return msg["content"].rstrip().endswith(termination_word)
    elif isinstance(msg.get("content"), list):
        for content in msg["content"]:
            if isinstance(content, dict) and "text" in content:
                return content["text"].rstrip().endswith(termination_word)
    return False

def always_terminate(msg) -> bool:
    return True


def custom_speaker_selection_func(last_speaker, groupchat):
    """Define a customized speaker selection function.
    A recommended way is to define a transition for each speaker in the groupchat.

    Parameters:
        - last_speaker: Agent
            The last speaker in the group chat.
        - groupchat: GroupChat
            The GroupChat object
    Return:
        Return one of the following:
        1. an `Agent` class, it must be one of the agents in the group chat.
        2. a string from ['auto', 'manual', 'random', 'round_robin'] to select a default method to use.
        3. None, which indicates the chat should be terminated.
    """
    pass


class AutogenAgentType(Enum):
    ConversableAgent = 1
    UserProxyAgent = 2
    AssistantAgent = 3
    GroupChatManager = 4
    

class Example(BaseModel):
    format: Literal["text", "json", "markdown", "html"]
    content: str
    header: str = ""

class Task(BaseModel):
    name: str
    description: str
    expected_output: str
    
    def to_crewai_task(self, agent: Agent) -> 'CrewAITask':
        logger.info(f"Creating CrewAI Task {self.name}")
        return CrewAITask(description=self.description, expected_output=self.expected_output, agent=agent)

    # def __str__(self) -> str:

class Role(BaseModel):
    name: str
    description: str  # The CrewAI's goal of the role
    agent_system_message: str  # The CrewAI's background 
    examples: List[Example] = []
    tasks: List[Task] = []
    human_input_mode: Literal["ALWAYS", "NEVER", "TERMINATE"] = "TERMINATE"
    autogen_code_execution_config: dict = {}

    @classmethod
    def from_json_file(cls, file_path: str) -> 'Role':
        with open(file_path, 'r') as file:
            json_data = json.load(file)
        return cls(**json_data)
    
    @classmethod
    def from_yaml_file(cls, file_path: str) -> 'Role':
        with open(file_path, 'r') as file:
            yaml_data = yaml.safe_load(file)
        return cls(**yaml_data)
    
    def to_yaml(self):
        return yaml.dump(self.model_dump())

    def build_path(self, path, file_name, overwrite_existing, extension: str = "json"):
        if path is None:
            path = Path(self.__class__.__name__)
        if file_name is None:
            file_name = self.name
        if not file_name.endswith(f".{extension}"):
            file_name += f".{extension}"
        path = Path(path) / file_name
        
        # Create the directory if it does not exist
        path.parent.mkdir(parents=True, exist_ok=overwrite_existing)
        logger.info(f"Path {path.parent} created")
        
        return path

    def to_yaml_file(self,
                     path: Optional[Path] = None, 
                     file_name: Optional[str] = None, 
                     overwrite_existing: bool = True):
        
        path = self.build_path(path, file_name, overwrite_existing, extension="yaml")
        
        with open(path, 'w') as file:
            file.write(self.to_yaml())
        logger.info(f"{self.name} object written to yaml: {path}")
    
    def to_crewai_agent(self, 
                        verbose: bool = False,
                        allow_delegation: bool = False,
                        llm_config: Optional[dict] = None,
                        associate_tasks: bool = False) -> Union['Agent', Tuple['Agent', List['Task']]]:
        agent = Agent(
            role=self.name,
            goal=self.description,
            backstory=self.agent_system_message,
            llm=llm_config,
            verbose=verbose,
            allow_delegation=allow_delegation,
        )
        
        if not associate_tasks:
            return agent
        else:
            return agent, [task.to_crewai_task(agent) for task in self.tasks]
        

    def to_autogen_agent(self, 
                         name:str, 
                         type: AutogenAgentType, 
                         human_input_mode: Optional[Literal["ALWAYS", "NEVER", "TERMINATE"]] = None,
                         llm_config: Optional[dict] = None,
                         max_consecutive_auto_reply: Optional[int] = None,
                         group_chat: Optional[GroupChat] = None,
                         code_execution_config: Optional[Union[Dict, Literal[False]]] = False,
                         termination_function: Optional[Callable] = None) -> 'ConversableAgent':
        
        code_execution_config = code_execution_config if code_execution_config is not None else self.autogen_code_execution_config
        
        if termination_function is not None:
            termination_message_placeholder = lambda msg: termination_function(msg)
        else:
            termination_message_placeholder = None
        
        match type:
            case AutogenAgentType.ConversableAgent:
                return ConversableAgent(name=name, 
                                                system_message=self.agent_system_message,
                                                description=self.description,
                                                llm_config=llm_config,
                                                human_input_mode=human_input_mode if human_input_mode is not None else self.human_input_mode,
                                                max_consecutive_auto_reply=max_consecutive_auto_reply,
                                                is_termination_msg=termination_message_placeholder,
                                                code_execution_config=code_execution_config
                )
            case AutogenAgentType.AssistantAgent:
                return AssistantAgent(name=name, 
                                              system_message=self.agent_system_message, 
                                              description=self.description,
                                              llm_config=llm_config,
                                              human_input_mode=human_input_mode if human_input_mode is not None else self.human_input_mode,
                                              max_consecutive_auto_reply=max_consecutive_auto_reply,
                                              is_termination_msg=termination_message_placeholder,
                                              code_execution_config=code_execution_config,
                )
            case AutogenAgentType.UserProxyAgent:
                return UserProxyAgent(name=name, 
                                              system_message=self.agent_system_message, 
                                              description=self.description,
                                              llm_config=llm_config,
                                              human_input_mode=human_input_mode if human_input_mode is not None else self.human_input_mode,
                                              max_consecutive_auto_reply=max_consecutive_auto_reply,
                                              is_termination_msg=termination_message_placeholder,
                                              code_execution_config=code_execution_config,
                )
            case AutogenAgentType.GroupChatManager:
                if group_chat is None:
                    raise ValueError("Group chat is required for GroupChatManager")
                return GroupChatManager(name=name, 
                                                groupchat=group_chat,
                )
            case _:
                raise ValueError(f"Invalid agent type: {type}")

    def get_examples_as_str(self) -> str:
        examples_str = ""
        
        for example in self.examples:
            if example.header != "":
                examples_str += f"{example.header}\n\n"
            if example.format == "json":
                examples_str += f"{json.dumps(json.loads(example.content), indent=4)}\n\n"
            else:
                examples_str += f"{example.content}\n\n"

        return examples_str

class Persona(BaseModel):
    name: str
    roles: Dict[str, Role]
    
    @classmethod
    def from_json_file(cls, file_path: str) -> 'Persona':
        with open(file_path, 'r') as file:
            json_data = json.load(file)
        return cls(**json_data)

    def build_path(self, path, file_name, overwrite_existing, extension: str = "json"):
        if path is None:
            path = Path(self.__class__.__name__)
        if file_name is None:
            file_name = self.name
        if not file_name.endswith(f".{extension}"):
            file_name += f".{extension}"
        path = Path(path) / file_name
        
        # Create the directory if it does not exist
        path.parent.mkdir(parents=True, exist_ok=overwrite_existing)
        logger.info(f"Path {path.parent} created")
        
        return path

    def save_as_json(self, 
                     path: Optional[Path] = None, 
                     file_name: Optional[str] = None, 
                     overwrite_existing: bool = True):
        
        path = self.build_path(path, file_name, overwrite_existing)
        
        with open(path, 'w') as file:
            file.write(self.model_dump_json(indent=4))
            
        logger.info(f"JSON object written: {path}")
        
    @classmethod
    def from_yaml_file(cls, file_path: str) -> 'Persona':
        with open(file_path, 'r') as file:
            yaml_data = yaml.safe_load(file)
        
        logger.info(f"YAML data:\n{yaml_data}")
        
        return cls(**yaml_data)
        
    def to_yaml(self):
        return yaml.dump(self.model_dump())

    def to_yaml_file(self,
                     path: Optional[Path] = None, 
                     file_name: Optional[str] = None, 
                     overwrite_existing: bool = True):
        
        path = self.build_path(path, file_name, overwrite_existing, extension="yaml")
        
        with open(path, 'w') as file:
            file.write(self.to_yaml())
        logger.info(f"{self.name} object written to yaml: {path}")

    def get_role(self, name: str) -> Role:
        try:
            return self.roles[name]
        except KeyError:
            raise ValueError(f"Role {name} not found in Persona {self.name}")

        
    def get_roles(self) -> List[str]:
        return list(self.roles.keys())


    def role_to_autogen_agent(self,
                              role_name: str,
                              type: AutogenAgentType, 
                              human_input_mode:Literal["ALWAYS", "NEVER", "TERMINATE"] = "NEVER",
                              llm_config: Optional[dict] = None,
                              group_chat: Optional[GroupChat] = None,
                              code_execution_config: Optional[Union[Dict, Literal[False]]] = False,
                              termination_function: Optional[Callable] = None) -> 'ConversableAgent':
        
        return self.roles[role_name].to_autogen_agent(name=f"{role_name}", 
                                                       type=type, 
                                                       human_input_mode=human_input_mode,
                                                       llm_config=llm_config,
                                                       group_chat=group_chat,
                                                       code_execution_config=code_execution_config,
                                                       termination_function=termination_function)

    def __str__(self) -> str:
        def format_task(task_name: str, task: Task, indent: int) -> str:
            indent_str = " " * indent
            task_str = f"{indent_str}Task: {task_name}\n"
            task_str += f"{indent_str}Description: {task.description}\n"
            task_str += f"{indent_str}Expected output: {task.expected_output}\n"
            return task_str
        
        def format_role(role_name: str, role: Role, indent: int) -> str:
            indent_str = " " * indent
            role_str = f"{indent_str}Role: {role_name}\n"
            role_str += f"{indent_str}Description: {role.description}\n"
            role_str += f"{indent_str}Agent System Message: {role.agent_system_message}\n"
            role_str += f"{indent_str}Autogen Code Execution Config: {role.autogen_code_execution_config}\n"
            for task in role.tasks:
                role_str += format_task(task.name, task, indent + 2)
            return role_str

        def format_persona(persona: Persona, indent: int) -> str:
            indent_str = " " * indent
            persona_str = f"\n{indent_str}Persona: {persona.name}\n"
            for role_name, role in persona.roles.items():
                persona_str += format_role(role_name, role, indent + 2)
            return persona_str

        return format_persona(self, 0)
    
    
class Application(BaseModel):
    name: str
    roles: Dict[str, Role]
    
    @classmethod
    def from_json_file(cls, file_path: str) -> 'Application':
        with open(file_path, 'r') as file:
            json_data = json.load(file)
        return cls(**json_data) 
